import { ISerializeObject, ISerializer } from './interface'
import { isPlainObject } from '../utils'

class UnhandledTypeError extends TypeError {
  constructor(serializer: ISerializer<unknown>, value: any) {
    super(`unhandled type: '${typeof value}' for serializer ${serializer.constructor.name}`)
  }
}

function serializeObject<T, V extends object, S extends ISerializeObject<T>>(serializer: S, obj: V): T {
  for (const key in obj) {
    const value = obj[key]
    serializer.serializeEntry(key, value)
  }

  return serializer.end()
}

function serializeClass<T, V extends object, S extends ISerializer<T>>(serializer: S, value: V): T {
  const name = value.constructor.name
  const ser = serializer.serializeClass(name)
  return serializeObject(ser, value)
}

export function serialize<T, V, S extends ISerializer<T>>(serializer: S, value: V): T {
  switch (typeof value) {
    case 'string': return serializer.serializeString(value)
    case 'number': return serializer.serializeNumber(value)
    case 'bigint': return serializer.serializeBigInt(value)
    case 'boolean': return serializer.serializeBoolean(value)
    case 'symbol': return serializer.serializeSymbol(value)
    case 'undefined': return serializer.serializeNull()
    case 'object':
      switch (true) {
        case value == null: return serializer.serializeNull()
        case !isPlainObject(value): return serializeClass(serializer, value)
        default: return serializeObject(serializer.serializeObject(), value)
      }
    default: throw new UnhandledTypeError(serializer, value)
  }
}

